from __future__ import annotations

import llama_cpp
from llama_cpp.server.model import LlamaProxy
from llama_cpp.server.settings import ModelSettings
from llama_cpp.server.__main__ import main

from empower_functions.chat_handler import EmpowerFunctionsCompletionHandler
import json
from typing import Optional, Union, Dict
import llama_cpp
from llama_cpp.server.model import (
    LlamaProxy,
)
from llama_cpp.server.settings import (
    ModelSettings,
)
from empower_functions.monkey_patch.app import patch_app

# Monkey pacthing the LlamaProxy class


def load_llama_from_model_settings(settings: ModelSettings) -> llama_cpp.Llama:
    chat_handler = None
    if settings.chat_format == "empower-functions":
        chat_handler = EmpowerFunctionsCompletionHandler()
    elif settings.chat_format == "llava-1-5":
        assert settings.clip_model_path is not None, "clip model not found"
        if settings.hf_model_repo_id is not None:
            chat_handler = (
                llama_cpp.llama_chat_format.Llava15ChatHandler.from_pretrained(
                    repo_id=settings.hf_model_repo_id,
                    filename=settings.clip_model_path,
                    verbose=settings.verbose,
                )
            )
        else:
            chat_handler = llama_cpp.llama_chat_format.Llava15ChatHandler(
                clip_model_path=settings.clip_model_path, verbose=settings.verbose
            )
    elif settings.chat_format == "obsidian":
        assert settings.clip_model_path is not None, "clip model not found"
        if settings.hf_model_repo_id is not None:
            chat_handler = (
                llama_cpp.llama_chat_format.ObsidianChatHandler.from_pretrained(
                    repo_id=settings.hf_model_repo_id,
                    filename=settings.clip_model_path,
                    verbose=settings.verbose,
                )
            )
        else:
            chat_handler = llama_cpp.llama_chat_format.ObsidianChatHandler(
                clip_model_path=settings.clip_model_path, verbose=settings.verbose
            )
    elif settings.chat_format == "llava-1-6":
        assert settings.clip_model_path is not None, "clip model not found"
        if settings.hf_model_repo_id is not None:
            chat_handler = (
                llama_cpp.llama_chat_format.Llava16ChatHandler.from_pretrained(
                    repo_id=settings.hf_model_repo_id,
                    filename=settings.clip_model_path,
                    verbose=settings.verbose,
                )
            )
        else:
            chat_handler = llama_cpp.llama_chat_format.Llava16ChatHandler(
                clip_model_path=settings.clip_model_path, verbose=settings.verbose
            )
    elif settings.chat_format == "moondream":
        assert settings.clip_model_path is not None, "clip model not found"
        if settings.hf_model_repo_id is not None:
            chat_handler = (
                llama_cpp.llama_chat_format.MoondreamChatHandler.from_pretrained(
                    repo_id=settings.hf_model_repo_id,
                    filename=settings.clip_model_path,
                    verbose=settings.verbose,
                )
            )
        else:
            chat_handler = llama_cpp.llama_chat_format.MoondreamChatHandler(
                clip_model_path=settings.clip_model_path, verbose=settings.verbose
            )
    elif settings.chat_format == "nanollava":
        assert settings.clip_model_path is not None, "clip model not found"
        if settings.hf_model_repo_id is not None:
            chat_handler = (
                llama_cpp.llama_chat_format.NanoLlavaChatHandler.from_pretrained(
                    repo_id=settings.hf_model_repo_id,
                    filename=settings.clip_model_path,
                    verbose=settings.verbose,
                )
            )
        else:
            chat_handler = llama_cpp.llama_chat_format.NanoLlavaChatHandler(
                clip_model_path=settings.clip_model_path, verbose=settings.verbose
            )
    elif settings.chat_format == "llama-3-vision-alpha":
        assert settings.clip_model_path is not None, "clip model not found"
        if settings.hf_model_repo_id is not None:
            chat_handler = (
                llama_cpp.llama_chat_format.Llama3VisionAlpha.from_pretrained(
                    repo_id=settings.hf_model_repo_id,
                    filename=settings.clip_model_path,
                    verbose=settings.verbose,
                )
            )
        else:
            chat_handler = llama_cpp.llama_chat_format.Llama3VisionAlpha(
                clip_model_path=settings.clip_model_path, verbose=settings.verbose
            )
    elif settings.chat_format == "hf-autotokenizer":
        assert (
            settings.hf_pretrained_model_name_or_path is not None
        ), "hf_pretrained_model_name_or_path must be set for hf-autotokenizer"
        chat_handler = (
            llama_cpp.llama_chat_format.hf_autotokenizer_to_chat_completion_handler(
                settings.hf_pretrained_model_name_or_path
            )
        )
    elif settings.chat_format == "hf-tokenizer-config":
        assert (
            settings.hf_tokenizer_config_path is not None
        ), "hf_tokenizer_config_path must be set for hf-tokenizer-config"
        chat_handler = llama_cpp.llama_chat_format.hf_tokenizer_config_to_chat_completion_handler(
            json.load(open(settings.hf_tokenizer_config_path))
        )

    tokenizer: Optional[llama_cpp.BaseLlamaTokenizer] = None
    if settings.hf_pretrained_model_name_or_path is not None:
        tokenizer = llama_tokenizer.LlamaHFTokenizer.from_pretrained(
            settings.hf_pretrained_model_name_or_path
        )

    draft_model = None
    if settings.draft_model is not None:
        draft_model = llama_speculative.LlamaPromptLookupDecoding(
            num_pred_tokens=settings.draft_model_num_pred_tokens
        )

    kv_overrides: Optional[Dict[str, Union[bool, int, float]]] = None
    if settings.kv_overrides is not None:
        assert isinstance(settings.kv_overrides, list)
        kv_overrides = {}
        for kv in settings.kv_overrides:
            key, value = kv.split("=")
            if ":" in value:
                value_type, value = value.split(":")
                if value_type == "bool":
                    kv_overrides[key] = value.lower() in ["true", "1"]
                elif value_type == "int":
                    kv_overrides[key] = int(value)
                elif value_type == "float":
                    kv_overrides[key] = float(value)
                else:
                    raise ValueError(f"Unknown value type {value_type}")

    import functools

    kwargs = {}

    if settings.hf_model_repo_id is not None:
        create_fn = functools.partial(
            llama_cpp.Llama.from_pretrained,
            repo_id=settings.hf_model_repo_id,
            filename=settings.model,
        )
    else:
        create_fn = llama_cpp.Llama
        kwargs["model_path"] = settings.model

    _model = create_fn(
        **kwargs,
        # Model Params
        n_gpu_layers=settings.n_gpu_layers,
        main_gpu=settings.main_gpu,
        tensor_split=settings.tensor_split,
        vocab_only=settings.vocab_only,
        use_mmap=settings.use_mmap,
        use_mlock=settings.use_mlock,
        kv_overrides=kv_overrides,
        # Context Params
        seed=settings.seed,
        n_ctx=settings.n_ctx,
        n_batch=settings.n_batch,
        n_threads=settings.n_threads,
        n_threads_batch=settings.n_threads_batch,
        rope_scaling_type=settings.rope_scaling_type,
        rope_freq_base=settings.rope_freq_base,
        rope_freq_scale=settings.rope_freq_scale,
        yarn_ext_factor=settings.yarn_ext_factor,
        yarn_attn_factor=settings.yarn_attn_factor,
        yarn_beta_fast=settings.yarn_beta_fast,
        yarn_beta_slow=settings.yarn_beta_slow,
        yarn_orig_ctx=settings.yarn_orig_ctx,
        mul_mat_q=settings.mul_mat_q,
        logits_all=settings.logits_all,
        embedding=settings.embedding,
        offload_kqv=settings.offload_kqv,
        flash_attn=settings.flash_attn,
        # Sampling Params
        last_n_tokens_size=settings.last_n_tokens_size,
        # LoRA Params
        lora_base=settings.lora_base,
        lora_path=settings.lora_path,
        # Backend Params
        numa=settings.numa,
        # Chat Format Params
        chat_format=settings.chat_format,
        chat_handler=chat_handler,
        # Speculative Decoding
        draft_model=draft_model,
        # KV Cache Quantization
        type_k=settings.type_k,
        type_v=settings.type_v,
        # Tokenizer
        tokenizer=tokenizer,
        # Misc
        verbose=settings.verbose,
    )
    if settings.cache:
        if settings.cache_type == "disk":
            if settings.verbose:
                print(f"Using disk cache with size {settings.cache_size}")
            cache = llama_cpp.LlamaDiskCache(
                capacity_bytes=settings.cache_size)
        else:
            if settings.verbose:
                print(f"Using ram cache with size {settings.cache_size}")
            cache = llama_cpp.LlamaRAMCache(capacity_bytes=settings.cache_size)
        _model.set_cache(cache)
    return _model


LlamaProxy.load_llama_from_model_settings = staticmethod(
    load_llama_from_model_settings)

if __name__ == "__main__":
    patch_app()

    main()



import json

SYSTEM_INSTRUCTION = "In this environment you have access to a set of functions defined in the JSON format you can use to address user's requests, use them if needed."


def _check_functions_def(functions_def):
    """Check if the functions definition is valid."""

    for function in functions_def:
        if 'name' not in function:
            raise 'Function name must be provided'
        if 'description' not in function:
            raise 'Function description must be provided'
        if 'parameters' not in function:
            raise 'Function parameters must be provided'

        parameters = function['parameters']
        if 'type' not in parameters:
            raise 'Function parameters type must be provided'
        if parameters['type'] != 'object':
            raise 'Function parameters type must be object'
        if 'properties' not in parameters:
            raise 'Function parameters properties must be provided'

        properties = parameters['properties']
        if not isinstance(properties, dict):
            raise 'Function parameters properties must be an object'
        if 'required' in parameters and not isinstance(parameters['required'], list):
            raise 'Function parameters required must be an array'


def _check_and_merge_messages(messages):
    """Check if the messages are valid."""
    if len(messages) == 0:
        raise 'Messages cannot be empty'

    first_message = messages[0]
    updated_messages = []
    if first_message['role'] == 'system':
        messages = messages[1:]
        updated_messages.append(first_message)

    if len(messages) == 0:
        raise 'At least user message must be provided'

    previous_role = ''
    for message in messages:
        if 'role' not in message or message['role'] not in ['user', 'assistant', 'tool']:
            raise Exception('Invalid role')

        if message['role'] == 'tool':
            if 'content' not in message:
                raise Exception(
                    '"content" must be provided for message with role "tool"')
            if 'tool_call_id' not in message or not message['tool_call_id']:
                raise Exception(
                    '"tool_call_id" must be provided for message with role "tool"')

            try:
                content = json.loads(message['content'])
            except:
                raise Exception(
                    'Content of a message with role "tool" must be a valid JSON string')

            if previous_role == 'tool':
                updated_messages[-1]['content'].append({
                    "value": content,
                    'tool_call_id': message['tool_call_id']
                })
            else:
                updated_messages.append({
                    "role": "tool",
                    "content": [{
                        "value": content,
                        'tool_call_id': message['tool_call_id']
                    }]
                })

        elif message['role'] == 'user':
            if 'content' not in message:
                raise Exception(
                    '"content" must be provided for message with role "user"')
            if previous_role == 'user':
                updated_messages[-1]['content'] += "\n\n" + message['content']
            else:
                updated_messages.append(message)
        elif message['role'] == 'assistant':
            if previous_role == 'assistant':
                raise Exception(
                    'Consecutive assistant messages are not allowed')

            if 'content' not in message and 'tool_calls' not in message:
                raise Exception(
                    'Either "content" or "tool_calls" must be provided message with role "assistant"')

            if 'tool_calls' in message:
                tool_calls = message['tool_calls']
                if not isinstance(tool_calls, list):
                    raise Exception('Tool calls must be an array')
                for tool_call in tool_calls:
                    if not isinstance(tool_call, dict):
                        raise Exception('Tool call must be an object')
                    if 'id' not in tool_call:
                        raise Exception(
                            '"id" must be provided for tool call')
                    if 'function' not in tool_call:
                        raise Exception(
                            '"function" must be provided for tool call')

                    function = tool_call['function']
                    if not isinstance(function, dict):
                        raise Exception('Tool call function must be an object')
                    if 'name' not in function:
                        raise Exception(
                            '"name" must be provided for the function in tool call')
                    if 'arguments' not in function:
                        raise Exception(
                            '"arguments" must be provided for the function in tool call')

            updated_messages.append(message)

        previous_role = message['role']

    return updated_messages


def prompt_messages(messages, functions_def, include_thinking=False):
    if not functions_def:
        functions_def = []

    if len(functions_def) == 0 and include_thinking:
        raise Exception(
            'Currently thinking mode is only supported with tools. Please provide functions_def to enable thinking mode.')

    _check_functions_def(functions_def)
    messages = _check_and_merge_messages(messages)

    system_instruction = SYSTEM_INSTRUCTION
    if include_thinking:
        system_instruction += "\nMake sure to include your thinking inside < thinking > </thinking > before response."

    first_user_message = messages[0]
    starting_index = 1
    if messages[0]['role'] == 'system':
        system_instruction = messages[0]['content']
        first_user_message = messages[1]
        starting_index = 2

    if len(functions_def) == 0:
        prompted_messages = [first_user_message]
    else:
        prompted_messages = [{'role': 'user', 'content': (
            system_instruction
            + "\n"
            + "Functions:\n"
            + json.dumps(functions_def, indent=2, ensure_ascii=False)
            + "\n\n"
            + "User Message:\n"
            + first_user_message['content']
        )}]

    for message in messages[starting_index:]:
        if message['role'] == 'tool':
            prompted_messages.append({
                'role': 'user',
                'content': '<r>' + json.dumps(message['content'], indent=2, ensure_ascii=True)
            })
        elif message['role'] == 'user':
            prompted_messages.append({
                'role': 'user',
                'content': '<u>' + message['content']
            })
        elif message['role'] == 'assistant':
            if 'content' in message and message['content'] and len(message['content']) > 0:
                prompted_messages.append({
                    'role': 'assistant',
                    'content': '<c>' + message['content']
                })
            else:
                functions = [tool_call['function']
                             for tool_call in message['tool_calls']]
                prompted_messages.append({
                    'role': 'assistant', 'content': '<f>' + json.dumps(functions, indent=2, ensure_ascii=False)
                })

    return prompted_messages



import json

from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    Protocol,
    cast,
)

import jinja2
from jinja2.sandbox import ImmutableSandboxedEnvironment

import llama_cpp.llama as llama
import llama_cpp.llama_types as llama_types
from llama_cpp.llama_chat_format import LlamaChatCompletionHandler
from empower_functions.prompt import prompt_messages
import traceback


class EmpowerFunctionsCompletionHandler(LlamaChatCompletionHandler):
    def __call__(
        self,
        llama: llama.Llama,
        messages: List[llama_types.ChatCompletionRequestMessage],
        functions: Optional[List[llama_types.ChatCompletionFunction]] = None,
        function_call: Optional[llama_types.ChatCompletionRequestFunctionCall] = None,
        tools: Optional[List[llama_types.ChatCompletionTool]] = None,
        tool_choice: Optional[llama_types.ChatCompletionToolChoiceOption] = None,
        temperature: float = 0.2,
        top_p: float = 0.95,
        top_k: int = 40,
        min_p: float = 0.05,
        typical_p: float = 1.0,
        stream: bool = False,
        stop: Optional[Union[str, List[str]]] = [],
        response_format: Optional[
            llama_types.ChatCompletionRequestResponseFormat
        ] = None,
        max_tokens: Optional[int] = None,
        presence_penalty: float = 0.0,
        frequency_penalty: float = 0.0,
        repeat_penalty: float = 1.1,
        tfs_z: float = 1.0,
        mirostat_mode: int = 0,
        mirostat_tau: float = 5.0,
        mirostat_eta: float = 0.1,
        model: Optional[str] = None,
        logits_processor: Optional[llama.LogitsProcessorList] = None,
        grammar: Optional[llama.LlamaGrammar] = None,
        logprobs: Optional[bool] = None,
        top_logprobs: Optional[int] = None,
        **kwargs,  # type: ignore
    ) -> Union[
        llama_types.CreateChatCompletionResponse,
        Iterator[llama_types.CreateChatCompletionStreamResponse],
    ]:
        template = "{% set loop_messages = messages %}{% for message in loop_messages %}{% set content = '<|start_header_id|>' + message['role'] + '<|end_header_id|>\n\n'+ message['content'] | trim + '<|eot_id|>' %}{% if loop.index0 == 0 %}{% set content = '<|begin_of_text|>' + content %}{% endif %}{{ content }}{% endfor %}{% if add_generation_prompt %}{{ '<|start_header_id|>assistant<|end_header_id|>\n\n' }}{% endif %}"
        template_renderer = ImmutableSandboxedEnvironment(
            autoescape=False,
            undefined=jinja2.StrictUndefined,
        ).from_string(template)

        # Convert legacy function_call to tool_choice
        if function_call is not None:
            if isinstance(function_call, str) and (
                function_call == "none" or function_call == "auto"
            ):
                tool_choice = function_call
            if isinstance(function_call, dict) and "name" in function_call:
                tool_choice = {
                    "type": "function",
                    "function": {
                        "name": function_call["name"],
                    },
                }
        if not tool_choice:
            tool_choice = "auto"

        assert tool_choice != "any"

        if functions is None:
            functions = [tool.get("function") for tool in tools]
        stop = (
            [stop, "<|eot_id|>"]
            if isinstance(stop, str)
            else stop + ["<|eot_id|>"] if stop else ["<|eot_id|>"]
        )

        if tool_choice == "none":
            functions = []

        include_thinking = False
        if "include_thinking" in kwargs:
            include_thinking = kwargs["include_thinking"]
        prompted_messages = prompt_messages(
            messages, functions, include_thinking=include_thinking
        )
        prompt = template_renderer.render(
            messages=prompted_messages, add_generation_prompt=True
        )

        # Case 1: No tool choice by user
        generated = llama.create_completion(
            prompt=prompt,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            min_p=min_p,
            typical_p=typical_p,
            stream=stream,
            stop=stop,
            max_tokens=max_tokens,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            repeat_penalty=repeat_penalty,
            tfs_z=tfs_z,
            mirostat_mode=mirostat_mode,
            mirostat_tau=mirostat_tau,
            mirostat_eta=mirostat_eta,
            model=model,
            logits_processor=logits_processor,
            grammar=grammar,
            logprobs=top_logprobs if logprobs else None,
        )
        thinking = None
        content = None

        (content, thinking) = _separate_thinking_if_present(
            generated["choices"][0]["text"]
        )
        if content.startswith("<f>"):
            generated["choices"][0]["text"] = content
            return _convert_completion_to_chat_function(
                completion_or_chunks=generated,
                thinking=thinking,
            )
        elif content.startswith("<c>"):
            generated["choices"][0]["text"] = thinking + content[3:]
            return _convert_completion_to_chat(generated, stream=stream)

        return _convert_completion_to_chat(generated, stream=stream)


def _convert_completion_to_chat(
    completion_or_chunks: Union[
        llama_types.CreateCompletionResponse,
        Iterator[llama_types.CreateCompletionStreamResponse],
    ],
    stream: bool = False,
) -> Union[
    llama_types.CreateChatCompletionResponse, Iterator[llama_types.ChatCompletionChunk]
]:
    if stream:
        # type: ignore
        chunks: Iterator[llama_types.CreateCompletionStreamResponse] = completion_or_chunks
        return _convert_text_completion_chunks_to_chat(chunks)
    else:
        completion: llama_types.Completion = completion_or_chunks  # type: ignore
        return _convert_text_completion_to_chat(completion)


def _maybe_json_dumps(value: Any) -> str:
    if isinstance(value, str):
        return value
    return json.dumps(value)


def _convert_completion_to_chat_function(
    completion_or_chunks: llama_types.CreateCompletionResponse,
    thinking: Optional[str] = None,
):
    completion: llama_types.CreateCompletionResponse = completion_or_chunks  # type: ignore
    assert "usage" in completion
    # tool_id = "call_" + "_0_" + tool_name + "_" + completion["id"]
    # TODO: Fix for legacy function calls
    json_object = json.loads(completion["choices"][0]["text"][3:])

    tool_calls = [
        {
            "id": "call_"
            + "_0_"
            + tool["name"]
            + "_"
            + completion["id"]
            + "_"
            + str(i),
            "type": "function",
            "function": {
                "name": tool["name"],
                "arguments": _maybe_json_dumps(tool["arguments"]),
            },
        }
        for (i, tool) in enumerate(json_object)
    ]

    json_object = json.loads(completion["choices"][0]["text"][3:])

    chat_completion: llama_types.CreateChatCompletionResponse = {
        "id": "chat" + completion["id"],
        "object": "chat.completion",
        "created": completion["created"],
        "model": completion["model"],
        "choices": [
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": thinking,
                    "tool_calls": tool_calls,
                },
                "logprobs": completion["choices"][0]["logprobs"],
                "finish_reason": "tool_calls",
            }
        ],
        "usage": completion["usage"],
    }
    return chat_completion


def _convert_text_completion_chunks_to_chat(
    chunks: Iterator[llama_types.CreateCompletionStreamResponse],
) -> Iterator[llama_types.ChatCompletionChunk]:
    for i, chunk in enumerate(chunks):
        if i == 0:
            yield {
                "id": "chat" + chunk["id"],
                "model": chunk["model"],
                "created": chunk["created"],
                "object": "chat.completion.chunk",
                "choices": [
                    {
                        "index": 0,
                        "delta": {
                            "role": "assistant",
                        },
                        "logprobs": None,
                        "finish_reason": None,
                    }
                ],
            }
        yield {
            "id": "chat" + chunk["id"],
            "model": chunk["model"],
            "created": chunk["created"],
            "object": "chat.completion.chunk",
            "choices": [
                {
                    "index": 0,
                    "delta": (
                        {
                            "content": chunk["choices"][0]["text"],
                        }
                        if chunk["choices"][0]["finish_reason"] is None
                        else {}
                    ),
                    "logprobs": chunk["choices"][0]["logprobs"],
                    "finish_reason": chunk["choices"][0]["finish_reason"],
                }
            ],
        }


def _convert_completion_to_chat(
    completion_or_chunks: Union[
        llama_types.CreateCompletionResponse,
        Iterator[llama_types.CreateCompletionStreamResponse],
    ],
    stream: bool = False,
) -> Union[
    llama_types.CreateChatCompletionResponse, Iterator[llama_types.ChatCompletionChunk]
]:
    if stream:
        # type: ignore
        chunks: Iterator[llama_types.CreateCompletionStreamResponse] = completion_or_chunks
        return _convert_text_completion_chunks_to_chat(chunks)
    else:
        completion: llama_types.Completion = completion_or_chunks  # type: ignore
        return _convert_text_completion_to_chat(completion)


def _convert_text_completion_to_chat(
    completion: llama_types.Completion,
) -> llama_types.ChatCompletion:
    assert "usage" in completion
    return {
        "id": "chat" + completion["id"],
        "object": "chat.completion",
        "created": completion["created"],
        "model": completion["model"],
        "choices": [
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": completion["choices"][0]["text"],
                },
                "logprobs": completion["choices"][0]["logprobs"],
                "finish_reason": completion["choices"][0]["finish_reason"],
            }
        ],
        "usage": completion["usage"],
    }


def _convert_text_completion_chunks_to_chat(
    chunks: Iterator[llama_types.CreateCompletionStreamResponse],
) -> Iterator[llama_types.ChatCompletionChunk]:
    for i, chunk in enumerate(chunks):
        if i == 0:
            yield {
                "id": "chat" + chunk["id"],
                "model": chunk["model"],
                "created": chunk["created"],
                "object": "chat.completion.chunk",
                "choices": [
                    {
                        "index": 0,
                        "delta": {
                            "role": "assistant",
                        },
                        "logprobs": None,
                        "finish_reason": None,
                    }
                ],
            }
        yield {
            "id": "chat" + chunk["id"],
            "model": chunk["model"],
            "created": chunk["created"],
            "object": "chat.completion.chunk",
            "choices": [
                {
                    "index": 0,
                    "delta": (
                        {
                            "content": chunk["choices"][0]["text"],
                        }
                        if chunk["choices"][0]["finish_reason"] is None
                        else {}
                    ),
                    "logprobs": chunk["choices"][0]["logprobs"],
                    "finish_reason": chunk["choices"][0]["finish_reason"],
                }
            ],
        }


def _separate_thinking_if_present(text):
    tag = "</thinking>"
    tag_position = text.find(tag)

    if tag_position != -1:
        # Split the string into two parts
        part1 = text[: tag_position + len(tag)]
        part2 = text[tag_position + len(tag):]
        return part2, part1
    else:
        return text, None



from setuptools import setup, find_packages
# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "empower_functions/README.md").read_text()

setup(
    name='empower_functions',  # This is the package name
    version='0.1.4',
    packages=find_packages(),  # This will find the `empower_functions` directory
    install_requires=[
        "jinja2",
        "llama-cpp-python[server]",
        "pydantic"
        # Add your dependencies here
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
)
